"""Asset tracking functionality helps track assets by hostname for 
the purpose of easily identifying rogue devices or devices without
all of an enterprises security tools.
"""

__version__ = "0.1.0"
